==> lms-0.0.x-dev.tar.gz: lms/opigno_learning_path/opigno_learning_path.post_update.php <==
<?php

/**
 * @file
 * Post update functions for opigno_learning_path.
 */

use Drupal\Core\Site\Settings;
use Drupal\group\Entity\Group;
use Drupal\group\Entity\GroupContent;
use Drupal\group\Entity\GroupContentType;
use Drupal\group\Entity\GroupRole;
use Drupal\opigno_learning_path\Entity\LPStatus;
use Drupal\opigno_module\Entity\OpignoModule;
use Drupal\user\Entity\User;

/**
 * Enable membership requests for learning paths.
 */
function opigno_learning_path_post_update_enable_membership_requests(&$sandbox) {
  $database = \Drupal::database();
  if (!isset($sandbox['progress'])) {
    $count = $database->query("SELECT COUNT(*) FROM {opigno_learning_path_group_user_status}")->fetchField();
    $sandbox['max'] = $count;
    // Ensure configuration exists.
    $sandbox['progress'] = 0;
    $values = [
      'status' => TRUE,
      'id' => 'group_content_type_c7a4e16345d82',
      'label' =>  'Learning Path: Group membership request',
      'description' => 'Adds users as requesters for the group.',
      'group_type' => 'learning_path',
      'content_plugin' => 'group_membership_request',
      'plugin_config' => [
         'group_cardinality' => 0,
         'entity_cardinality' => 1,
         'use_creation_wizard' => false,
       ],
    ];
    $group_content = GroupContentType::load($values['id']);
    if ($group_content) {
      $group_content->set('plugin_config', $values['plugin_config']);
    }
    else {
      $group_content = GroupContentType::create($values);
    }
    $group_content->save();

    $role = GroupRole::load('learning_path-outsider');
    $role->grantPermission('request membership');
    $role->save();


    $role = GroupRole::load('learning_path-admin');
    $role->grantPermission('administer membership requests');
    $role->save();

    $role = GroupRole::load('learning_path-user_manager');
    $role->grantPermission('administer membership requests');
    $role->save();
  }
  if (!$sandbox['max']) {
    return;
  }
  $membership_request_manager = Drupal::service('grequest.membership_request_manager');
  $result = $database->queryRange("SELECT * FROM {opigno_learning_path_group_user_status} ORDER BY id ASC", $sandbox['progress'], 50);
  foreach ($result as $record) {
    $sandbox['progress']++;
    $group = Group::load($record->gid);
    $user = User::load($record->uid);
    $membership = GroupContent::load($record->mid);
    // Only process the record if all three entities exist.
    if ($group && $user && $membership) {
      // For 'blocked' status create a 'rejected' membership request.
      if ($record->status === '3') {
        $membership->delete();
        \Drupal::service('group.membership_loader')->resetUserStaticCache($user);
        \Drupal::service('group.membership_loader')->resetGroupStaticCache($group);
        // Check for duplicate memberships. If there's a duplicate, this one is
        // invalid and trust that we'll process the duplicate later.
        if (!$group->getMember($user)) {
          $membership_request = $membership_request_manager->create($group, $user);
          $membership_request->reject();
        }
      }
      elseif ($record->status === '2') {
        $membership->delete();
        \Drupal::service('group.membership_loader')->resetUserStaticCache($user);
        \Drupal::service('group.membership_loader')->resetGroupStaticCache($group);
        // Check for duplicate memberships. If there's a duplicate, this one is
        // invalid and trust that we'll process the duplicate later.
        if (!$group->getMember($user)) {
          $membership_request = $membership_request_manager->create($group, $user);
          $membership_request->save();
        }
      }
    }
  }

  $sandbox['#finished'] = $sandbox['progress'] / $sandbox['max'];
  return sprintf('Processed %d of %d membership status records.', $sandbox['progress'], $sandbox['max']);
}

/**
 * Remove multiple entries from opigno_latest_group_activity.
 */
function opigno_learning_path_post_update_remove_latest_activity_multiples(&$sandbox) {
  if (!isset($sandbox['list'])) {
    $sandbox['list'] = \Drupal::database()->query("SELECT uid, training FROM {opigno_latest_group_activity} GROUP BY uid, training HAVING COUNT(*) > 1")->fetchAll(PDO::FETCH_ASSOC);
    $sandbox['total'] = count($sandbox['list']);
    $sandbox['current'] = 0;
  }

  if (array_key_exists($sandbox['current'], $sandbox['list'])) {
    $item = $sandbox['list'][$sandbox['current']];
    $latest_activities = \Drupal::entityTypeManager()->getStorage('opigno_latest_group_activity')->loadByProperties([
      'uid' => $item['uid'],
      'training' => $item['training'],
    ]);

    $latest_ts = 0;
    $latest_id = 0;
    foreach ($latest_activities as $id => $latest_activity) {
      if ($latest_activity->getTimestamp() >= $latest_ts) {
        $latest_id = $id;
      }
    }

    foreach ($latest_activities as $id => $latest_activity) {
      if ($latest_id === $id) {
        continue;
      }

      $latest_activity->delete();
    }

    $sandbox['current']++;
    $sandbox['#finished'] = $sandbox['current'] / $sandbox['total'];
  }

}

/**
 * AFLLAB-25 (from EL-288): Consolidate roles.
 */
function opigno_learning_path_post_update_consolidate_roles(&$sandbox) {
  $mapping = [
    'learning_path' => [
      [
        ['content_manager', 'learning_path_mana'],
        ['admin'],
      ],
      [['student'], []],
      [['coach'], ['facilitator']],
    ],
    'class' => [
      [['student'], []],
    ],
  ];

  $gs = \Drupal::entityTypeManager()->getStorage('group');
  if (!isset($sandbox['group_ids'])) {
    $query = $gs->getQuery();
    $sandbox['group_ids'] = $query
      ->condition('type', array_keys($mapping), 'IN')
      ->accessCheck(FALSE)
      ->execute();
    $sandbox['group_ids'] = array_values($sandbox['group_ids']);
    $sandbox['total'] = count($sandbox['group_ids']);
    $sandbox['progress'] = 0;
  }
  if (array_key_exists($sandbox['progress'], $sandbox['group_ids'])) {
    $group = $gs->load($sandbox['group_ids'][$sandbox['progress']]);

    $type_mapping = $mapping[$group->bundle()];
    foreach ($type_mapping as $conversion_item) {
      $to_remove = [];
      foreach ($conversion_item[0] as $role) {
        $to_remove[] = $group->bundle() . '-' . $role;
      }

      $to_add = [];
      foreach ($conversion_item[1] as $role) {
        $to_add[] = $group->bundle() . '-' . $role;
      }

      // Get members having any of the source roles.
      $members = $group->getMembers($to_remove);

      foreach ($members as $member) {
        $group_content_entity = $member->getGroupContent();
        $role_values = $group_content_entity->group_roles->getValue();
        // Remove roles to be consolidated first.
        $role_values = array_filter($role_values, fn($item) => !in_array($item['target_id'], $to_remove));

        // Add target roles if not already present.
        foreach ($to_add as $new_role) {
          if (!array_filter($role_values, fn($item) => $item['target_id'] === $new_role)) {
            $role_values[]['target_id'] = $new_role;
          }
        }
        $group_content_entity->group_roles->setValue($role_values);
        $group_content_entity->save();
      }

    }

    $sandbox['progress']++;
    $sandbox['#finished'] = $sandbox['progress'] / $sandbox['total'];
  }
}

/**
 * Add default class to learning paths.
 */
function opigno_learning_path_post_update_add_missing_classes(&$sandbox) {
  $group_storage = \Drupal::entityTypeManager()->getStorage('group');
  $group_content_storage = \Drupal::entityTypeManager()->getStorage('group_content');
  $groupHierarchyManager = \Drupal::service('ggroup.group_hierarchy_manager');

  if (!\array_key_exists('list', $sandbox)) {
    $sandbox['list'] = $group_storage
      ->getQuery()
      ->accessCheck(FALSE)
      ->condition('type', 'learning_path')
      ->execute();
    $sandbox['total'] = \count($sandbox['list']);
    $sandbox['progress'] = 0;
    $sandbox['added_classes'] = 0;
    $sandbox['moved_members'] = 0;
  }

  $id = array_pop($sandbox['list']);
  if ($id === NULL) {
    return;
  }

  $class = NULL;

  $learning_path = $group_storage->load($id);

  $classes = [];
  $subgroups = $groupHierarchyManager->getGroupSubgroups($id);
  foreach ($subgroups as $subgroup) {
    if ($subgroup->bundle() === 'opigno_class') {
      $classes[] = $subgroup;
    }
  }

  // Find members without roles that are not members of any classes.
  $default_class_additions = [];
  foreach ($learning_path->getMembers() as $membership) {
    $account = $membership->getUser();
    // Byproduct data cleanup - memberships with no user entities.
    if ($account === NULL) {
      $membership->getGroupContent()->delete();
      continue;
    }

    // If the user has learning path roles, it's an admin user,
    // no action required.
    if (\count($membership->getRoles()) !== 1) {
      continue;
    }

    $in_class = FALSE;
    $account = $membership->getUser();
    foreach ($classes as $class) {
      if ($class->getMember($account) !== FALSE) {
        $in_class = TRUE;
        continue;
      }
    }

    // Remove non admin LP membership.
    $learning_path->removeMember($account);
    // Add to default class queue if the account is not a member of a class.
    if (!$in_class) {
      $default_class_additions[] = $account;
    }
  }

  // Create a default class if there are no classes at all or if
  // there are users to move.
  if (count($classes) === 0 || count($default_class_additions) !== 0) {
    $class = $group_storage->create([
      'type' => 'opigno_class',
      'label' => t('@training (default class)', [
        '@training' => $learning_path->label(),
      ]),
      'uid' => $learning_path->getOwnerId(),
    ]);
    $class->save();
    $content = $group_content_storage->createForEntityInGroup($class, $learning_path, 'subgroup:opigno_class');
    $content->setSyncing(TRUE);
    $content->save();
    $content->setSyncing(FALSE);

    $sandbox['added_classes']++;
  }

  // Move memberships.
  if (count($default_class_additions) !== 0) {
    foreach ($default_class_additions as $account) {
      $content = $group_content_storage->createForEntityInGroup($account, $class, 'group_membership');
      $content->setSyncing(TRUE);
      $content->save();
      $content->setSyncing(FALSE);
      $sandbox['moved_members']++;
    }
  }

  $sandbox['progress']++;
  $sandbox['#finished'] = $sandbox['progress'] / $sandbox['total'];
  \Drupal::logger('opigno_learning_path')->notice('Checked %progress, added %added_classes classes, moved %moved_members members, %remaining learning paths remaining.', [
    '%progress' => $sandbox['progress'],
    '%added_classes' => $sandbox['added_classes'],
    '%moved_members' => $sandbox['moved_members'],
    '%remaining' => $sandbox['total'] - $sandbox['progress'],
  ]);
}

/**
 * AFLLAB-632: Fix failed learning path statuses.
 */
function opigno_learning_path_post_update_fix_failed(&$sandbox) {
  $lp_result_storage = \Drupal::entityTypeManager()->getStorage('user_lp_status');

  if (!\array_key_exists('list', $sandbox)) {
    $sandbox['list'] = $lp_result_storage
      ->getQuery()
      ->accessCheck(FALSE)
      ->condition('status', LPStatus::STATUS_FAILED)
      ->execute();
    $sandbox['total'] = \count($sandbox['list']);
    $sandbox['progress'] = 0;
    $sandbox['updated'] = 0;
  }

  $training_manager = \Drupal::service('opigno_learning_path.training_manager');

  for ($i = 0; $i < Settings::get('entity_update_batch_size', 50); $i++) {
    if (\count($sandbox['list']) === 0) {
      return;
    }

    $id = array_pop($sandbox['list']);
    $lp_status = $lp_result_storage->load($id);
    $training_manager->updateLpStatus($lp_status);
    if ($lp_status->getStatus() !== LPStatus::STATUS_FAILED) {
      $sandbox['updated']++;
    }

    $sandbox['progress']++;
    $sandbox['#finished'] = $sandbox['progress'] / $sandbox['total'];
    \Drupal::logger('opigno_learning_path')->notice('Updated @count LPStatus entities. @remaining remaining to check.', [
      '@count' => $sandbox['updated'],
      '@remaining' => $sandbox['total'] - $sandbox['progress'],
    ]);
  }
}

/**
 * AFLLAB-632: Fix failed learning path statuses.
 */
function opigno_learning_path_post_update_last_activity_ts(&$sandbox) {
  $lp_result_storage = \Drupal::entityTypeManager()->getStorage('user_lp_status');

  if (!\array_key_exists('list', $sandbox)) {
    $sandbox['list'] = $lp_result_storage
      ->getQuery()
      ->accessCheck(FALSE)
      ->execute();
    $sandbox['total'] = \count($sandbox['list']);
    $sandbox['progress'] = 0;
  }

  $latest_activity_storage = \Drupal::entityTypeManager()->getStorage('opigno_latest_group_activity');

  if (\count($sandbox['list']) === 0) {
    return;
  }

  $id = array_pop($sandbox['list']);
  $lp_status = $lp_result_storage->load($id);

  // Copy-paste from former _aft_opigno_learning_path_user_status_data().
  // Ok, modified slightly actually.
  $result = $latest_activity_storage->getAggregateQuery()
    ->condition('uid', $lp_status->get('uid')->target_id)
    ->condition('training', $lp_status->get('gid')->target_id)
    ->accessCheck(FALSE)
    ->aggregate('timestamp', 'MAX')
    ->execute();
  $timestamp = is_null($result[0]['timestamp_max']) ? 0 : (int) $result[0]['timestamp_max'];

  $group = $lp_status->getTraining();
  $user = $lp_status->getUser();

  // Yes, there were cases of both actually.
  // @todo: Borken reference LPStatus cleanup implementation.
  if ($group === NULL || $user === NULL) {
    $lp_status->delete();
  }
  else {
    $module_ids = opigno_learning_path_get_modules_ids_by_group($group);
    $modules = OpignoModule::loadMultiple($module_ids);

    foreach ($modules as $module) {
      $attempts = $module->getModuleAttempts($user, NULL, NULL, TRUE, $group->id());
      foreach ($attempts as $attempt) {
        $attempt_finished_ts = (int) $attempt->get('finished')->value;
        if ($timestamp < $attempt_finished_ts) {
          $timestamp = $attempt_finished_ts;
        }
      }
    }
    if ($timestamp === 0) {
      $timestamp = (int) $lp_status->get('started')->value;
    }
    $lp_status->set('last_activity_ts', $timestamp);
    $lp_status->save();
  }

  $sandbox['progress']++;
  $sandbox['#finished'] = $sandbox['progress'] / $sandbox['total'];
  \Drupal::logger('opigno_learning_path')->notice('Updated @count LPStatus entities. @remaining remaining.', [
    '@count' => $sandbox['progress'],
    '@remaining' => $sandbox['total'] - $sandbox['progress'],
  ]);
}

==> lms-0.2.x-dev.tar.gz: lms/opigno_learning_path/opigno_learning_path.post_update.php <==
<?php

/**
 * @file
 * Post update functions for opigno_learning_path.
 */

use Drupal\Core\Site\Settings;
use Drupal\group\Entity\Group;
use Drupal\group\Entity\GroupRelationship;
use Drupal\group\Entity\GroupRelationshipType;
use Drupal\group\Entity\GroupRole;
use Drupal\opigno_learning_path\Entity\LPStatus;
use Drupal\opigno_module\Entity\OpignoModule;
use Drupal\user\Entity\User;

/**
 * Enable membership requests for learning paths.
 */
function opigno_learning_path_post_update_enable_membership_requests(&$sandbox) {
  $database = \Drupal::database();
  if (!isset($sandbox['progress'])) {
    $count = $database->query("SELECT COUNT(*) FROM {opigno_learning_path_group_user_status}")->fetchField();
    $sandbox['max'] = $count;
    // Ensure configuration exists.
    $sandbox['progress'] = 0;
    $values = [
      'status' => TRUE,
      'id' => 'group_content_type_c7a4e16345d82',
      'label' =>  'Learning Path: Group membership request',
      'description' => 'Adds users as requesters for the group.',
      'group_type' => 'learning_path',
      'content_plugin' => 'group_membership_request',
      'plugin_config' => [
         'group_cardinality' => 0,
         'entity_cardinality' => 1,
         'use_creation_wizard' => false,
       ],
    ];
    $group_content = GroupRelationshipType::load($values['id']);
    if ($group_content) {
      $group_content->set('plugin_config', $values['plugin_config']);
    }
    else {
      $group_content = GroupRelationshipType::create($values);
    }
    $group_content->save();

    $role = GroupRole::load('learning_path-outsider');
    $role->grantPermission('request membership');
    $role->save();


    $role = GroupRole::load('learning_path-admin');
    $role->grantPermission('administer membership requests');
    $role->save();

    $role = GroupRole::load('learning_path-user_manager');
    $role->grantPermission('administer membership requests');
    $role->save();
  }
  if (!$sandbox['max']) {
    return;
  }
  $membership_request_manager = Drupal::service('grequest.membership_request_manager');
  $result = $database->queryRange("SELECT * FROM {opigno_learning_path_group_user_status} ORDER BY id ASC", $sandbox['progress'], 50);
  foreach ($result as $record) {
    $sandbox['progress']++;
    $group = Group::load($record->gid);
    $user = User::load($record->uid);
    $membership = GroupRelationship::load($record->mid);
    // Only process the record if all three entities exist.
    if ($group && $user && $membership) {
      // For 'blocked' status create a 'rejected' membership request.
      if ($record->status === '3') {
        $membership->delete();
        \Drupal::service('group.membership_loader')->resetUserStaticCache($user);
        \Drupal::service('group.membership_loader')->resetGroupStaticCache($group);
        // Check for duplicate memberships. If there's a duplicate, this one is
        // invalid and trust that we'll process the duplicate later.
        if (!$group->getMember($user)) {
          $membership_request = $membership_request_manager->create($group, $user);
          $membership_request->reject();
        }
      }
      elseif ($record->status === '2') {
        $membership->delete();
        \Drupal::service('group.membership_loader')->resetUserStaticCache($user);
        \Drupal::service('group.membership_loader')->resetGroupStaticCache($group);
        // Check for duplicate memberships. If there's a duplicate, this one is
        // invalid and trust that we'll process the duplicate later.
        if (!$group->getMember($user)) {
          $membership_request = $membership_request_manager->create($group, $user);
          $membership_request->save();
        }
      }
    }
  }

  $sandbox['#finished'] = $sandbox['progress'] / $sandbox['max'];
  return sprintf('Processed %d of %d membership status records.', $sandbox['progress'], $sandbox['max']);
}

/**
 * Remove multiple entries from opigno_latest_group_activity.
 */
function opigno_learning_path_post_update_remove_latest_activity_multiples(&$sandbox) {
  if (!isset($sandbox['list'])) {
    $sandbox['list'] = \Drupal::database()->query("SELECT uid, training FROM {opigno_latest_group_activity} GROUP BY uid, training HAVING COUNT(*) > 1")->fetchAll(PDO::FETCH_ASSOC);
    $sandbox['total'] = count($sandbox['list']);
    $sandbox['current'] = 0;
  }

  if (array_key_exists($sandbox['current'], $sandbox['list'])) {
    $item = $sandbox['list'][$sandbox['current']];
    $latest_activities = \Drupal::entityTypeManager()->getStorage('opigno_latest_group_activity')->loadByProperties([
      'uid' => $item['uid'],
      'training' => $item['training'],
    ]);

    $latest_ts = 0;
    $latest_id = 0;
    foreach ($latest_activities as $id => $latest_activity) {
      if ($latest_activity->getTimestamp() >= $latest_ts) {
        $latest_id = $id;
      }
    }

    foreach ($latest_activities as $id => $latest_activity) {
      if ($latest_id === $id) {
        continue;
      }

      $latest_activity->delete();
    }

    $sandbox['current']++;
    $sandbox['#finished'] = $sandbox['current'] / $sandbox['total'];
  }

}

/**
 * AFLLAB-25 (from EL-288): Consolidate roles.
 */
function opigno_learning_path_post_update_consolidate_roles(&$sandbox) {
  $mapping = [
    'learning_path' => [
      [
        ['content_manager', 'learning_path_mana'],
        ['admin'],
      ],
      [['student'], []],
      [['coach'], ['facilitator']],
    ],
    'class' => [
      [['student'], []],
    ],
  ];

  $gs = \Drupal::entityTypeManager()->getStorage('group');
  if (!isset($sandbox['group_ids'])) {
    $query = $gs->getQuery();
    $sandbox['group_ids'] = $query
      ->condition('type', array_keys($mapping), 'IN')
      ->accessCheck(FALSE)
      ->execute();
    $sandbox['group_ids'] = array_values($sandbox['group_ids']);
    $sandbox['total'] = count($sandbox['group_ids']);
    $sandbox['progress'] = 0;
  }
  if (array_key_exists($sandbox['progress'], $sandbox['group_ids'])) {
    $group = $gs->load($sandbox['group_ids'][$sandbox['progress']]);

    $type_mapping = $mapping[$group->bundle()];
    foreach ($type_mapping as $conversion_item) {
      $to_remove = [];
      foreach ($conversion_item[0] as $role) {
        $to_remove[] = $group->bundle() . '-' . $role;
      }

      $to_add = [];
      foreach ($conversion_item[1] as $role) {
        $to_add[] = $group->bundle() . '-' . $role;
      }

      // Get members having any of the source roles.
      $members = $group->getMembers($to_remove);

      foreach ($members as $member) {
        $group_content_entity = $member->getGroupRelationship();
        $role_values = $group_content_entity->group_roles->getValue();
        // Remove roles to be consolidated first.
        $role_values = array_filter($role_values, fn($item) => !in_array($item['target_id'], $to_remove));

        // Add target roles if not already present.
        foreach ($to_add as $new_role) {
          if (!array_filter($role_values, fn($item) => $item['target_id'] === $new_role)) {
            $role_values[]['target_id'] = $new_role;
          }
        }
        $group_content_entity->group_roles->setValue($role_values);
        $group_content_entity->save();
      }

    }

    $sandbox['progress']++;
    $sandbox['#finished'] = $sandbox['progress'] / $sandbox['total'];
  }
}

/**
 * AFLLAB-632: Fix failed learning path statuses.
 */
function opigno_learning_path_post_update_fix_failed(&$sandbox) {
  $lp_result_storage = \Drupal::entityTypeManager()->getStorage('user_lp_status');

  if (!\array_key_exists('list', $sandbox)) {
    $sandbox['list'] = $lp_result_storage
      ->getQuery()
      ->accessCheck(FALSE)
      ->condition('status', LPStatus::STATUS_FAILED)
      ->execute();
    $sandbox['total'] = \count($sandbox['list']);
    $sandbox['progress'] = 0;
    $sandbox['updated'] = 0;
  }

  $training_manager = \Drupal::service('opigno_learning_path.training_manager');

  for ($i = 0; $i < Settings::get('entity_update_batch_size', 50); $i++) {
    if (\count($sandbox['list']) === 0) {
      return;
    }

    $id = array_pop($sandbox['list']);
    $lp_status = $lp_result_storage->load($id);
    $training_manager->updateLpStatus($lp_status);
    if ($lp_status->getStatus() !== LPStatus::STATUS_FAILED) {
      $sandbox['updated']++;
    }

    $sandbox['progress']++;
    $sandbox['#finished'] = $sandbox['progress'] / $sandbox['total'];
    \Drupal::logger('opigno_learning_path')->notice('Updated @count LPStatus entities. @remaining remaining to check.', [
      '@count' => $sandbox['updated'],
      '@remaining' => $sandbox['total'] - $sandbox['progress'],
    ]);
  }
}

/**
 * AFLLAB-632: Fix failed learning path statuses.
 */
function opigno_learning_path_post_update_last_activity_ts(&$sandbox) {
  $lp_result_storage = \Drupal::entityTypeManager()->getStorage('user_lp_status');

  if (!\array_key_exists('list', $sandbox)) {
    $sandbox['list'] = $lp_result_storage
      ->getQuery()
      ->accessCheck(FALSE)
      ->execute();
    $sandbox['total'] = \count($sandbox['list']);
    $sandbox['progress'] = 0;
  }

  $latest_activity_storage = \Drupal::entityTypeManager()->getStorage('opigno_latest_group_activity');

  if (\count($sandbox['list']) === 0) {
    return;
  }

  $id = array_pop($sandbox['list']);
  $lp_status = $lp_result_storage->load($id);

  // Copy-paste from former _aft_opigno_learning_path_user_status_data().
  // Ok, modified slightly actually.
  $result = $latest_activity_storage->getAggregateQuery()
    ->condition('uid', $lp_status->get('uid')->target_id)
    ->condition('training', $lp_status->get('gid')->target_id)
    ->accessCheck(FALSE)
    ->aggregate('timestamp', 'MAX')
    ->execute();
  $timestamp = is_null($result[0]['timestamp_max']) ? 0 : (int) $result[0]['timestamp_max'];

  $group = $lp_status->getTraining();
  $user = $lp_status->getUser();

  // Yes, there were cases of both actually.
  // @todo: Borken reference LPStatus cleanup implementation.
  if ($group === NULL || $user === NULL) {
    $lp_status->delete();
  }
  else {
    $module_ids = opigno_learning_path_get_modules_ids_by_group($group);
    $modules = OpignoModule::loadMultiple($module_ids);

    foreach ($modules as $module) {
      $attempts = $module->getModuleAttempts($user, NULL, NULL, TRUE, $group->id());
      foreach ($attempts as $attempt) {
        $attempt_finished_ts = (int) $attempt->get('finished')->value;
        if ($timestamp < $attempt_finished_ts) {
          $timestamp = $attempt_finished_ts;
        }
      }
    }
    if ($timestamp === 0) {
      $timestamp = (int) $lp_status->get('started')->value;
    }
    $lp_status->set('last_activity_ts', $timestamp);
    $lp_status->save();
  }

  $sandbox['progress']++;
  $sandbox['#finished'] = $sandbox['progress'] / $sandbox['total'];
  \Drupal::logger('opigno_learning_path')->notice('Updated @count LPStatus entities. @remaining remaining.', [
    '@count' => $sandbox['progress'],
    '@remaining' => $sandbox['total'] - $sandbox['progress'],
  ]);
}

/**
 * Set one current LPStatus for every LP and user.
 */
function opigno_learning_path_post_update_set_current_lp_status(&$sandbox) {
  $lp_status_storage = \Drupal::entityTypeManager()->getStorage('user_lp_status');

  if (!\array_key_exists('list', $sandbox)) {
    $sandbox['list'] = $lp_status_storage
      ->getQuery()
      ->accessCheck(FALSE)
      ->sort('started', 'ASC')
      ->execute();
    $sandbox['total'] = \count($sandbox['list']);
    $sandbox['current_statuses'] = [];
    $sandbox['progress'] = 0;
    $sandbox['updated'] = 0;
  }

  if (\count($sandbox['list']) === 0) {
    return;
  }

  $id = array_pop($sandbox['list']);
  $lp_status = $lp_status_storage->load($id);
  $gid = $lp_status->get('gid')->target_id;
  $uid = $lp_status->get('uid')->target_id;

  if (
    !array_key_exists($gid, $sandbox['current_statuses']) ||
    !array_key_exists($uid, $sandbox['current_statuses'][$gid])
  ) {
    $current = (bool) $lp_status->get('current')->value;
    if (!$current) {
      $lp_status->set('current', TRUE);
      $lp_status->save();
    }
    $sandbox['current_statuses'][$gid][$uid] = TRUE;

    $sandbox['updated']++;
  }
  else {
    $lp_status->set('current', FALSE);
    $lp_status->save();
  }
  \Drupal::logger('opigno_learning_path')->notice('Updated @count LPStatus entities. @remaining remaining to check.', [
    '@count' => $sandbox['updated'],
    '@remaining' => $sandbox['total'] - $sandbox['progress'],
  ]);

  $sandbox['progress']++;
  $sandbox['#finished'] = $sandbox['progress'] / $sandbox['total'];
}

/**
 * Add default class to learning paths.
 */
function opigno_learning_path_post_update_add_missing_classes(&$sandbox) {
  $group_storage = \Drupal::entityTypeManager()->getStorage('group');
  $group_content_storage = \Drupal::entityTypeManager()->getStorage('group_content');
  $groupHierarchyManager = \Drupal::service('ggroup.group_hierarchy_manager');

  if (!\array_key_exists('list', $sandbox)) {
    $sandbox['list'] = $group_storage
      ->getQuery()
      ->accessCheck(FALSE)
      ->condition('type', 'learning_path')
      ->execute();
    $sandbox['total'] = \count($sandbox['list']);
    $sandbox['progress'] = 0;
  }

  $id = array_pop($sandbox['list']);
  if ($id === NULL) {
    return;
  }

  $class = NULL;

  $learning_path = $group_storage->load($id);

  $classes = [];
  $subgroups = $groupHierarchyManager->getGroupSubgroups($id);
  foreach ($subgroups as $subgroup) {
    if ($subgroup->bundle() === 'opigno_class') {
      $classes[] = $subgroup;
    }
  }

  // Find members without roles that are not members of any classes.
  $default_class_additions = [];
  $removed = 0;
  $new_members_count = 0;
  foreach ($learning_path->getMembers() as $membership) {
    $account = $membership->getUser();
    // Byproduct data cleanup - memberships with no user entities.
    if ($account === NULL) {
      $membership->getGroupRelationship()->delete();
      continue;
    }

    // If the user has learning path roles, it's an admin user,
    // no action required.
    if (\count($membership->getRoles(FALSE)) !== 0) {
      continue;
    }

    $in_class = FALSE;
    foreach ($classes as $class) {
      if ($class->getMember($account) !== FALSE) {
        $in_class = TRUE;
        continue;
      }
    }

    // Remove non admin LP membership.
    $learning_path->removeMember($account);
    $removed++;
    // Add to default class queue if the account is not a member of a class.
    if (!$in_class) {
      $default_class_additions[] = $account;
    }
  }

  // Create a default class if there are no classes at all or if
  // there are users to move.
  if (count($classes) === 0 || count($default_class_additions) !== 0) {
    // Generate a Class.
    $class = $group_storage->create([
      'type' => 'opigno_class',
      'label' => t('Class @class_name', [
        // Generate a class name based on the learning path. Using the learning
        // path name could make group listings hard to scan. Using the learning
        // path ID could create confusion with the class's own numeric ID. To
        // avoid this and also avoid hash collisions, add 100 to the learning
        // path ID then base36 encode and upper-case it. This gives us a
        // human-readable short hash which is superficially similar to the
        // kind of class names that are often used in schools (1A, DF1, etc.).
        '@class_name' => strtoupper(base_convert((string) ($learning_path->id()) + 100, 10, 36)),
      ]),
      'uid' => $learning_path->getOwnerId(),
    ]);
    $class->save();
    $content = $group_content_storage->createForEntityInGroup($class, $learning_path, 'subgroup:opigno_class');
    $content->setSyncing(TRUE);
    $content->save();
    $content->setSyncing(FALSE);

    // Add memberships.
    if (count($default_class_additions) !== 0) {
      foreach ($default_class_additions as $account) {
        if ($class->getMember($account) !== FALSE) {
          continue;
        }
        $content = $group_content_storage->createForEntityInGroup($account, $class, 'group_membership', [
          'uid' => $learning_path->getOwnerId(),
        ]);
        $content->setSyncing(TRUE);
        $content->save();
        $content->setSyncing(FALSE);
      }
    }
  }

  $sandbox['progress']++;
  $sandbox['#finished'] = $sandbox['progress'] / $sandbox['total'];
  \Drupal::logger('opigno_learning_path')->notice('Processed: %learning_path, added %count class members, removed %removed LP members. %remaining to process.', [
    '%learning_path' => $learning_path->label() . '(' . $learning_path->id() . ')',
    '%count' => count($default_class_additions),
    '%removed' => $removed,
    '%remaining' => $sandbox['total'] - $sandbox['progress'],
  ]);
}
